@@ -17,12 +17,19 @@
 #          Ralph Bean  <rbean@redhat.com>
 
 
+import fnmatch
 import os
 import six
 import sys
 import json as JSON
 from collections import defaultdict
 
+try:
+    from collections import OrderedDict
+except ImportError:
+    from ordereddict import OrderedDict
+
+from kitchen.iterutils import iterate
 from moksha.common.lib.helpers import appconfig
 
 # Look in the current directory for egg-info
@@ -189,22 +196,60 @@ class MokshaHub(object):
     def consume_stomp_message(self, message):
         from moksha.hub.reactor import reactor
 
-        topic = message['headers'].get('destination')
+        headers = message['headers']
+
+        # Prefer "original-destination"
+        topic = headers.get('original-destination')
+
+        # ...but if it is not present, fall back to "destination"
+        if not topic:
+            topic = headers.get('destination')
+
         if not topic:
             log.debug("Got message without a topic: %r" % message)
             return
 
+        # https://stomp.github.io/stomp-specification-1.2.html#Value_Encoding
+        if asbool(self.config.get('stomp_unescape_headers', True)):
+            replacements = OrderedDict([
+                ('\\r', '\r'),
+                ('\\n', '\n'),
+                ('\c', ':'),
+                ('\\\\', '\\'),
+            ])
+            for key in headers:
+                for old, new in replacements.items():
+                    headers[key] = headers[key].replace(old, new)
+
         # FIXME: only do this if the consumer wants it `jsonified`
         try:
-            body = JSON.loads(message['body'])
+            if message['body']:
+                body = JSON.loads(message['body'])
+            else:
+                body = {}
         except Exception as e:
-            log.warning('Cannot decode message from JSON: %s' % e)
+            log.warning('Cannot decode body from JSON: %s -> %r' % (e, message))
             #body = {}
             body = message['body']
 
         # feed all of our consumers
-        for callback in self.topics.get(topic, []):
-            reactor.callInThread(callback, {'body': body, 'topic': topic})
+        envelope = {'body': body, 'topic': topic, 'headers': headers}
+
+        handled = True
+
+        # Some consumers subscribe to topics directly
+        for pattern, callbacks in self.topics.items():
+            if fnmatch.fnmatch(topic, pattern):
+                for callback in callbacks:
+                    handled = handled and callback(envelope)
+
+        # Others subscribe to a queue composed of many topics..
+        subscription = headers.get('subscription')
+        if subscription != topic:
+            for callback in self.topics.get(subscription, []):
+                handled = handled and callback(envelope)
+
+        return handled
 
 
 class CentralMokshaHub(MokshaHub):
@@ -260,7 +305,7 @@ class CentralMokshaHub(MokshaHub):
                 self.transport.write(msg)
 
             def connectionLost(self, reason):
-                log.info("Lost Websocket connection.  Cleaning up.")
+                log.debug("Lost Websocket connection.  Cleaning up.")
                 self.moksha_hub.unsubscribe(self.send_to_ws)
 
             def dataReceived(self, data):
@@ -350,7 +395,10 @@ class CentralMokshaHub(MokshaHub):
             for consumer in pkg_resources.iter_entry_points('moksha.consumer'):
                 try:
                     c = consumer.load()
-                    self._consumers.append(c)
+                    try:
+                        self._consumers.extend(c) # assume to be collection
+                    except TypeError:
+                        self._consumers.append(c)
                 except Exception as e:
                     log.exception("Failed to load %r consumer." % consumer.name)
         else:
@@ -363,17 +411,17 @@ class CentralMokshaHub(MokshaHub):
                 if not getattr(c, "_initialized", None):
                     log.info("%s:%s not initialized." % (
                         c_class.__module__, c_class.__name__,))
+                    continue
 
                 self.consumers.append(c)
 
                 # This can be dynamically assigned during instantiation
-                topic = c.topic
+                for topic in iterate(c.topic):
+                    if topic not in self.topics:
+                        self.topics[topic] = []
 
-                if topic not in self.topics:
-                    self.topics[topic] = []
-
-                if c.consume not in self.topics[topic]:
-                    self.topics[topic].append(c.consume)
+                    if c._consume not in self.topics[topic]:
+                        self.topics[topic].append(c._consume)
 
             except Exception as e:
                 log.exception("Failed to init %r consumer." % c_class)
@@ -391,7 +439,10 @@ class CentralMokshaHub(MokshaHub):
             ], []):
                 try:
                     p = producer.load()
-                    self._producers.append(p)
+                    try:
+                        self._producers.extend(p) # assume to be collection
+                    except TypeError:
+                        self._producers.append(p)
                 except Exception as e:
                     log.exception("Failed to load %r producer." % producer.name)
         else:
