@@ -17,9 +17,14 @@
 #          Ralph Bean <rbean@redhat.com>
 
 try:
-    import stomper
+    # Try first to use modern stomp-1.1
+    import stomper.stomp_11 as stomper
 except ImportError:
-    pass
+    # Failing that, use whatever is available.
+    try:
+        import stomper
+    except ImportError:
+        pass
 
 import logging
 
@@ -46,7 +51,7 @@ class StompHubExtension(MessagingHubExtension, ClientFactory):
 
         uri = self.config.get('stomp_uri', None)
         if not uri:
-            port = self.config.get('stomp_port', 61613)
+            port = int(self.config.get('stomp_port', 61613))
             host = self.config.get('stomp_broker')
             uri = "%s:%i" % (host, port)
 
@@ -55,7 +60,7 @@ class StompHubExtension(MessagingHubExtension, ClientFactory):
         self.address_index = 0
 
         # An exponential delay used to back off if we keep failing.
-        self._delay = 0.1
+        self._delay = float(self.config.get('stomp_delay', '0.1'))
 
         self.username = self.config.get('stomp_user', 'guest')
         self.password = self.config.get('stomp_pass', 'guest')
@@ -63,6 +68,8 @@ class StompHubExtension(MessagingHubExtension, ClientFactory):
         self.key = self.config.get('stomp_ssl_key', None)
         self.crt = self.config.get('stomp_ssl_crt', None)
 
+        self.client_heartbeat = int(self.config.get('stomp_heartbeat', 0))
+
         self.connect(self.addresses[self.address_index], self.key, self.crt)
         super(StompHubExtension, self).__init__()
 
@@ -87,36 +94,73 @@ class StompHubExtension(MessagingHubExtension, ClientFactory):
             reactor.connectTCP(host, int(port), self)
 
     def buildProtocol(self, addr):
-        self._delay = 0.1
-        log.info("build protocol was called with %r" % addr)
+        self._delay = float(self.config.get('stomp_delay', '0.1'))
+        log.debug("build protocol was called with %r" % addr)
         self.proto = StompProtocol(self, self.username, self.password)
         return self.proto
 
-    def connected(self):
+    def connected(self, server_heartbeat):
+        if server_heartbeat and self.client_heartbeat:
+            interval = max(self.client_heartbeat, server_heartbeat)
+            log.debug("Heartbeat of %ims negotiated from (%i,%i); starting." % (
+                interval, self.client_heartbeat, server_heartbeat))
+            self.start_heartbeat(interval)
+        else:
+            log.debug("Skipping heartbeat initialization")
+
+        # Sometimes, a stomp consumer may wish to be subscribed to a queue
+        # which is composed of messages from many different topics.  In this
+        # case, the hub hands dispatching messages to the right consumers.
+        # This extension is only concerned with the queue, and negotiating that
+        # with the broker.
+        stomp_queue = self.config.get('stomp_queue', None)
+        if stomp_queue and self._topics and self.topics != [stomp_queue]:
+            log.info('Discarding consumer-specified topics in favor of '
+                     'stomp_queue=%s: %r' % (stomp_queue, self._topics))
+            # Overwrite the declarations of all of our consumers.
+            self._topics = [stomp_queue]
+
         for topic in self._topics:
             log.info('Subscribing to %s topic' % topic)
             self.subscribe(topic, callback=lambda msg: None)
-        self._topics = []
+
         for frame in self._frames:
-            log.info('Flushing queued frame')
+            log.debug('Flushing queued frame')
             self.proto.transport.write(frame.pack())
         self._frames = []
 
     def clientConnectionLost(self, connector, reason):
         log.info('Lost connection.  Reason: %s' % reason)
+        self.stop_heartbeat()
         self.failover()
 
     def clientConnectionFailed(self, connector, reason):
         log.error('Connection failed. Reason: %s' % reason)
+        self.stop_heartbeat()
         self.failover()
 
     def failover(self):
         self.address_index = (self.address_index + 1) % len(self.addresses)
         args = (self.addresses[self.address_index], self.key, self.crt,)
-        self._delay = self._delay * (1 + (2.0 / len(self.addresses)))
+        self._delay = min(60.0, self._delay * (1 + (2.0 / len(self.addresses))))
         log.info('(failover) reconnecting in %f seconds.' % self._delay)
         reactor.callLater(self._delay, self.connect, *args)
 
+    def start_heartbeat(self, interval):
+        self._heartbeat_enabled = True
+        reactor.callLater(interval / 1000.0, self.heartbeat, interval)
+
+    def heartbeat(self, interval):
+        if self._heartbeat_enabled:
+            self.proto.transport.write(chr(0x0A))  # Lub-dub
+            reactor.callLater(interval / 1000.0, self.heartbeat, interval)
+        else:
+            log.debug("(heartbeat stopped)")
+
+    def stop_heartbeat(self):
+        log.debug("stopping heartbeat")
+        self._heartbeat_enabled = False
+
     def send_message(self, topic, message, **headers):
         f = stomper.Frame()
         f.unpack(stomper.send(topic, message))
@@ -135,7 +179,7 @@ class StompHubExtension(MessagingHubExtension, ClientFactory):
             if topic not in self._topics:
                 self._topics.append(topic)
         else:
-            log.info("sending subscription to the protocol")
+            log.debug("sending subscription to the protocol")
             self.proto.subscribe(topic)
 
         super(StompHubExtension, self).subscribe(topic, callback)
