@@ -97,7 +97,7 @@ class ZMQHubExtension(BaseZMQHubExtension):
             try:
                 self.pub_socket.bind(endpoint)
             except zmq.ZMQError:
-                map(self.pub_socket.bind, hostname2ipaddr(endpoint))
+                map(self.pub_socket.bind, set(hostname2ipaddr(endpoint)))
 
         # Factory used to lazily produce subsequent subscribers
         self.twisted_zmq_factory = txzmq.ZmqFactory()
@@ -130,7 +130,7 @@ class ZMQHubExtension(BaseZMQHubExtension):
         required_attrs = ['zmq_publish_endpoints', 'zmq_subscribe_endpoints']
         for attr in required_attrs:
             if not config.get(attr, None):
-                log.warn("No '%s' set.  Are you sure?" % attr)
+                log.warning("No '%s' set.  Are you sure?" % attr)
                 continue
 
             endpoints = config[attr].split(',')
@@ -147,7 +147,7 @@ class ZMQHubExtension(BaseZMQHubExtension):
         try:
             self.pub_socket.send_multipart([topic, message])
         except zmq.ZMQError as e:
-            log.warn("Couldn't send message: %r" % e)
+            log.warning("Couldn't send message: %r" % e)
 
         super(ZMQHubExtension, self).send_message(topic, message, **headers)
 
@@ -178,10 +178,17 @@ class ZMQHubExtension(BaseZMQHubExtension):
                         self.connection_cls(
                             self.twisted_zmq_factory, endpoint)
                 except zmq.ZMQError as e:
-                    log.warn("Failed txzmq create on %r %r" % (endpoint, e))
+                    log.warning("Failed txzmq create on %r %r" % (endpoint, e))
                     continue
 
-                def chain_over_moksha_callbacks(_body, _topic):
+                def chain_over_moksha_callbacks(*parts):
+                    if len(parts) != 2:
+                        raise ValueError(
+                            "Moksha can only handle multipart messages with a "
+                            "topic followed by a body.  Got %r" % parts)
+
+                    _body, _topic = parts
+
                     if isinstance(_topic, six.binary_type):
                         _topic = _topic.decode('utf-8')
                     if isinstance(_body, six.binary_type):
